# Program written by Oleg N. Osychenko, Ph.D 
# e-mail: osychenko@gmail.com
# for the project MEDEAS
# Coding style definitions:
# (1) Follow PEP8 in style conventions, for example:
#       Class    | Camel case        | class StringManipulator():
#       Variable | Words joined by _ | joined_by_underscore = True
#       Function | Words joined by _ | def multi_word_name(words):
#       Constant | All uppercase     | SECRET_KEY = 42
# (2) No multiple statements on a single line, that is:
#      if this_is_bad_code: rewrite_code(); make_it_more_readable();
# (3) Convenient contractions for variables:
#     pop -- population 
#     hist -- historic
#     init -- initial
# (4) To be continued...   
################################################################ 
# The algorithmic base is the WoLim model based on "windows"
# that we call here "modules". Each "module" is a separate realm
# encompassing data and interactions logically bound. 
# Each module is an object inheriting from GenericModule class. 
# A module object has a set of internal variables (for its own use only)
# and exposed variables that it provides for the other objects.
################################################################
import numpy as np
import pandas as pd
#import scipy as sc
#import pylab as pl
#import csv
import openpyxl as opx
import initglobals
import math
import re

def get_xls_data(filename='inputs.xlsx',sheet='Cork',row_x_str='1',first_cell='A3'):
    """ 
    Implements vensim's GET XLS DATA('inputs.xlsx','sheet','argumentRowNumber','DatarowStartingCell',dataRowNumber,array_length),
    It reads the row until stumbles upon an empty cell (in X row), or reaches the maximum array length, defined by max_len
    Parameters
    ----------
    row_x_str: row index of arguments(x, usually years) as string
    first_cell: cell where data starts(included) as string
    """
    max_len = 256
    row_x = int(row_x_str)
    row_y = int(first_cell[1:])
    first_column = first_cell[0]

    wb = initglobals.wbinputs

    def letter_to_index(letter):
        """gives 1 for A, 2 for B, 3 for C, etc."""
        return ord(letter.upper()) - ord('A') + 1 
    
    def take_row():
        """Returns pandas Series object: 
        .index == argu (usually, years) 
        .values == data (the row of data until the nearest empty cell or until max_len length
        """

        def end_of_loop():
            raise StopIteration

        data = np.array(list(end_of_loop() if wb[sheet].cell(row=row_y,column=first_column+i).value is None\
            else wb[sheet].cell(row=row_y,column=first_column+i).value for i in range(max_len)))
        #may give errors if there's junk data inline 
        #argu = np.array([wb[sheet].cell(row=row_x,column=first_column+i).value for i in range(len(data))])
        argu = np.array(list(end_of_loop() if wb[sheet].cell(row=row_x,column=first_column+i).value is None\
            else wb[sheet].cell(row=row_x,column=first_column+i).value for i in range(max_len)))
        leng =  min(len(argu), len(data))
        for i in range(leng):
            try:
                arg = float(argu[i])
                dat = float(data[i])
            except ValueError:
                leng_new = i
                break
        argu = argu[:leng_new]
        data = data[:leng_new]
        return pd.Series(argu, data)

    return take_row()

def get_xls_data_by_workbook(wb=None,sheet='Cork',row_x_str='1',first_cell='A3', rows=None):
    """ 
    Implements vensim's GET XLS DATA('inputs.xlsx','sheet',argumentRowNumber,'DatarowStartingCell',dataRowNumber,array_length),
    It reads the row until stumbles upon an empty cell (in X row), or reaches the maximum array length, defined by max_len
    Parameters
    ----------
    row_x: row of arguments(x, usually years)
    first_column: where the data starts(this column included)
    row_y: data row (y)
    max_len: maximum array length
     """
    wb = initglobals.wbinputs
        #print("File workbook object not given")
        #return None

    max_len = 512

    def letter_to_index(letter):
        """ Gives 1 for A, 2 for B, 3 for C, etc."""
        return ord(letter) - ord('A') + 1 

    def int_indices_tuple(row_x_str=row_x_str, first_cell=first_cell):
        """ Converts input of the kind '3', 'AB36' from Excel function
        into a correct integer tuple representation: (3, 28, 36)
        (AB = 1*26+2 = 28).
        """
        cell = first_cell.upper()
        first_column = 0
        for i in range(len(cell)):
            a = cell[i]
            if ord('A') > ord(a) or ord(a) > ord('Z'):
                row_y = int(first_cell[i:])
                break
            else:
                first_column *= 26
                first_column += letter_to_index(a)
        return int(row_x_str), row_y, first_column

    row_x, row_y, first_column = int_indices_tuple()


    
    def take_row(row_y, length=None, argu=None):
        """ Returns pandas Series object: 
        .index == argu (usually, years) 
        .values == data (the row of data until the nearest empty cell or until max_len length
        """
        #print('entered take row', flush=True)
        def end_of_loop():
            raise StopIteration
        # stop when x cell is free
        if length is None:
            #may give errors if there's junk data inline
            data = np.array(list(end_of_loop() if wb[sheet].cell(row=row_y,column=first_column+i).value is None\
                else wb[sheet].cell(row=row_y,column=first_column+i).value for i in range(max_len)))
            #print(len(data))
            argu = np.array(list(end_of_loop() if wb[sheet].cell(row=row_x,column=first_column+i).value is None\
                else wb[sheet].cell(row=row_x,column=first_column+i).value for i in range(max_len)))
            leng =  min(len(argu), len(data))
            leng_new = leng
            for i in range(leng):
                try:
                    arg = float(argu[i])
                    dat = float(data[i])
                except ValueError:
                    leng_new = i
                    break
            argu = argu[:leng_new]
            data = data[:leng_new]
        else:
            data = np.array(list(wb[sheet].cell(row=row_y,column=first_column+i).value for i in range(length)))

        return pd.Series(data=data, index=argu)


    if rows is None:
        return take_row(row_y)
    else:
        try:
            rows_int = int(rows)
            rows_output = [take_row(row_y)]
            leng = len(rows_output[0])
            argu_x = np.array(rows_output[0].index)
            rows_output += [take_row(row_y + i, length=leng, argu=argu_x) for i in range(1, rows_int)]
            return rows_output
        except:
            print(f"Error: incorrect vector data reading definition in\n'{sheet}' '{row_x_str}' '{first_cell}'")
            raise



def get_xls_constants(filename='inputs.xlsx',sheet='Cork',cell='A1'):
    """ 
    Implements vensim's GET XLS CONSTANTS('inputs.xlsx','sheet','cell'),
    example GET XLS CONSTANTS('inputs.xlsx','MySheet','K41')
    Parameters
    ----------
    filename: excel file to read from, string
    sheet: string
    cell: string in the form LetterNumber, example: 'K41'
    """
    return initglobals.wbinputs[sheet][cell].value

