#!/usr/bin/env python
__author__ = "Oleg Osychenko, Roger Samsó"
__maintainer__ = "Roger Samsó"
__status__ = "Development"


import re
import pandas as pd
import numpy as np
from toolbox.logger.logger import log
import os
import xarray as xr
import xlsxwriter as xlw
import pysd
import getopt
import sys
import json
from openpyxl.utils import coordinate_from_string, column_index_from_string
from openpyxl import load_workbook

from pysd.py_backend import functions # do NOT delete this import

import time


def get_initial_user_input(config, run_params):
    """
    Ask for user input to update simulation parameters
    Get user input
    :param config: (dict) configuration parameters
    :param run_params: (dict) simulation parameters
    :return: updated conf  and run_params dictionaries according to user inputs
    """
    try:
        opts, args = getopt.getopt(sys.argv[1:], "hspt:i:f:r:x:")

        for opt, arg in opts:
            if opt == '-h':
                print('Run run.py module followed by any of the following options:\n'
                      '-i --> initial year of simulation (i.e. -i 1999) \n'
                      '-f --> final year of simulation (i.e. -f 2010) \n'
                      '-t --> time step of simulation (in years) (i.e. -t 0.05) \n'
                      '-r --> time step of simulation result (in years) (i.e. -r 1.0) \n'
                      '-s --> silent mode \n'
                      '-p --> opens the plot gui after simulation \n'
                      '-x --> scenario name (names should be the same as the input file tabs \n')

                sys.exit()

            elif opt in ("-s", "--silent"):
                config['silent'] = True
            elif opt in ("-i", "--initial_time"):
                run_params['initial_time'] = float(arg)
            elif opt in ("-f", "--final_time"):
                run_params['final_time'] = float(arg)
            elif opt in ("-t", "--time_step"):
                run_params['time_step'] = float(arg)
            elif opt in ("-r", "--return_timestep"):
                config['return_timestep'] = arg
            elif opt in ("-p", "--plot"):
                config['plot'] = True
            elif opt in ("-x", "--scen"):
                config['scenario_sheet'] = arg
            else:
                pass
    except getopt.GetoptError:
        log.error("Wrong parameter definition (run 'python run.py -h' to see the description of available parameters)")
        sys.exit()
    return config, run_params


def excel_to_df(file_name, sheet, skip):
    """
    Converts excel sheet into a pandas DataFrame
    :param file_name: (str) name of path to the file
    :param sheet: (str) sheet name
    :param skip: (int) number of rows to skip
    :return: (pandas DataFrame)
    """
    try:
        df = pd.read_excel(io=file_name, sheet_name=sheet, header=None, skiprows=skip)
    except FileNotFoundError:
        raise
    else:
        return df


def compare_dataframes(df1, df2):
    """
    Compares two dataframes and gets a list, which contains a list for every row
     in the dataframes, and each of these lists contain the column number of the
      values that are different between the two df's
    :param df1: (dataframe) output of transforming an excel into a dataframe (inputs.xlsx)
    :param df2: (dataframe) output of transforming an excel into a dataframe (inputs_default.xlsx)
    :return: (list) a list for every row in the excel file. Each list contains the column numbers of the values that
    changed from one excel to another
    """

    df = df1[df1.ne(df2)]
    # makes the list of updated cells in the form:
    # elem_id+2=row_index_in_xlsx, when row[elem_id]=[idx_1, idx_2,...]
    lst = []
    for _, row in df.iterrows():
        mask = row.notnull()
        lst += [row[mask].index.tolist()]

    # setting a tolerance value to identify true differences between cells
    tol = 1 / 100000

    diff_list = []
    for row, values in enumerate(lst, 0):
        if values:
            int_list = []
            for val in values:
                if isinstance(df1.iloc[row, val], (int, float)) and isinstance(df2.iloc[row, val], (int, float)):
                    if abs(df1.iloc[row, val] - df2.iloc[row, val]) > tol:
                        int_list.append(val)
                    else:
                        continue
                else:
                    continue
            diff_list.append(int_list)
        else:
            diff_list.append([])

    return diff_list


def find_updated_cells(config):

    """
    Parses the selected scenario sheet into a df, and the same for the default BAU sheet
    Compares the selected scenario with the BAU scenario and returns a list of tuples
    containing the updated cell name (e.g. B5) and the new value in that cell
    :param config: (dict) configuration file
    :return: list of tuples containing cell name and value
    """

    # Loading the default and current scenarios from the respective excel files into separate dataframes
    try:
        df_curr_scen = excel_to_df(os.path.join(config['folder'], config['inputs_file']), config['scenario_sheet'].upper(), [0])
        df_def_scen = excel_to_df(os.path.join(config['folder'], config['default_inputs_file']), 'BAU', [0])
    except:
        log.error('Wrong scenario name, please note that scenario names must be written all in capital leters')
        sys.exit(0)

    # comparing the two dataframes
    diffs = compare_dataframes(df_curr_scen, df_def_scen)

    cells_updated = []

    for idx_row, value in enumerate(diffs):
        if value:  # if cell_value =! [], i.e. there are updated cells in the row
            for idx_col in value:#idx_col = int(cell_value[0].split(' ')[-1]) - 1  # gives the column index, since cell_value='Something: column_index'. -1 since it's zero-based
                cell_str = str(xlw.utility.xl_col_to_name(idx_col)) + str(idx_row + 2)
                try:
                    cell_contains = df_curr_scen.iloc[idx_row, idx_col]
                except:
                    print('{} not modified'.format(cell_str))
                if isinstance(cell_contains, (int, float)):
                    cells_updated += [(cell_str, cell_contains)]

    return cells_updated


def read_file(filename, enc='utf-8'):
    """
    This function reads a txt file and encodes it into
    :param filename: (str) name of the file to open
    :param enc: (str) encoding, defaults to utf8
    :return: the file encoded in enc
    """

    try:
        with open(filename, 'r', encoding=enc) as f:
            return f.read()
    except FileNotFoundError:
        raise


def store_results_csv(result, config, return_cols):

    """
    Separates between xarrays and pandas Series in the output of pysd
    Transforms all xarrays into series, with names equal to the name of the
    function and it's index
    Concatenates all results in a dataframe (final_results) which is returned
    Stores a pickle of the final results

    :param result: output from simulation
    :return: final_results (pd.DataFrame)
    """

    # separate between variables represented as pandas Series and those that are xarrays
    xarrays_names = [name for name in result.columns if name in return_cols if any([isinstance(result[name].values[i], xr.core.dataarray.DataArray) for i in range(len(result.index))])]
    series_names = [x for x in result.columns if x in return_cols if x not in xarrays_names]

    results_xarrays = result[xarrays_names]
    results_df = result[series_names]

    df_xarrays = pd.DataFrame(index=results_xarrays.index)

    for col in results_xarrays.columns:
        b = pd.DataFrame()
        dates_nans = {}
        for date in results_xarrays.index:
            a = results_xarrays[col].loc[date]
            if isinstance(a, xr.core.dataarray.DataArray):
                new_data = a.to_dataframe(name=str(date)).T
                b = pd.concat([b, new_data], axis=0)
            elif isinstance(a, (int, np.ndarray)) or np.isnan(a):
                dates_nans[date] = a  # store the date and the value for
                # later use
            else:
                log.warning("Unknown type {} for {}".format(type(a), a))

        new_cols = []
        for x in b.columns:
            if isinstance(x, tuple):
                x = ', '.join(x)
            if '[' not in x:
                new_cols.append(col + '[' + x + ']')
            else:
                new_x = x.replace(']', ',') + x + ']'
                new_cols.append(new_x)

        b.columns = new_cols

        if dates_nans:
            for date, value in dates_nans.items():
                b.loc[str(date)] = [value]*len(b.columns)
                b.sort_index(inplace=True, ascending=True)

        df_xarrays = pd.concat([df_xarrays, b], axis=1)

    final_results = pd.concat([results_df, df_xarrays.dropna()], axis=1)

    # create file name for the results file
    results_filename = 'results_{}_{}_{}_{}'.format(
            config['scenario_sheet'],
            config['run_params']['initial_time'],
            config['run_params']['final_time'],
            config['run_params']['time_step'])

    file_path_csv = _results_naming(config, results_filename, 'csv')

    # storing results to csv file
    final_results.to_csv(file_path_csv)
    log.info('Simulation results file is located in {}'.format(file_path_csv))

    try:
        with open(os.path.join(config['folder'], 'last_output_conf.txt'), 'w') as f:
            f.write(';'.join(final_results.columns))
    except FileNotFoundError:
        raise

    if final_results.isna().any().any():
        nan_vars = final_results.columns[final_results.isna().any()].tolist()
        log.warning("There are NaN's in the timeseries of the following variables: {}, which might indicate convergence issues, try decreasing the time step".format(", ".join(nan_vars)))

    return final_results


def _results_naming(config, base_name, fmt):
    """This function renames old simulation results with the same name but adding
    _old in the end, or old_old if _old also exists

    :return (str) new file path
    """

    folder = config['folder']

    new_path = os.path.join(folder, '{}.{}'.format(base_name, fmt))
    old_path = os.path.join(folder, '{}_old.{}'.format(base_name, fmt))
    old_old_path = os.path.join(folder, '{}_old_old.{}'.format(base_name, fmt))

    if os.path.isfile(old_path):
        os.rename(old_path, old_old_path)
        log.info('Previous simulation output {0}_old.{1} moved to {0}_old_old.{1}'.format(base_name, fmt))

    if os.path.isfile(new_path):
        os.rename(new_path, old_path)
        log.info('Previous simulation output {0}.{1} moved to {0}_old.{1}'.format(base_name, fmt))

    return new_path


def update_ctts(config, cells_updated):

    """

    :param config:
    :param cells_updated:
    :return: two dictionaries:
        updated_params: (dict) contains only the updated parameters (key) and values that will be passed to the pysd.run method

    """

    # Todo the population and GDP cells should not be hardcoded here, but in a separate file

    with open(os.path.join(config['folder'], 'constants.json'), 'r') as ctts:
        const_cells = json.load(ctts)

    updated_params = {}

    population = False
    gdp = False

    is_cntt = []
    for item in cells_updated:
        if item[0] in const_cells.keys():
            updated_params[const_cells[item[0]]['new_name']] = item[1]
            is_cntt.append(item[0])
        elif re.match(r'([A-Z]){1,2}11', item[0]):  # if the value is in row 11
            population = True
            is_cntt.append(item[0])
        elif re.match(r'([A-Z]){1,2}6', item[0]):  # if the value is in row 6
            gdp = True
            is_cntt.append(item[0])

    if population or gdp: # or year_policy or year_policy_h or policy or policy_h: # if changes have been detected in the cells corresp. to gdp or population, the excel file is parsed
        raw_excel = pd.ExcelFile(os.path.join(config['folder'],'inputs.xlsx'))
        parsed_excel_sheet = raw_excel.parse(sheet_name=config['scenario_sheet'].upper(), header=None, index_col=None)

    if population:
        time_pop = parsed_excel_sheet.iloc[9, 4:90].tolist()
        vals_pop = parsed_excel_sheet.iloc[10, 4:90].tolist()
        new_pop = "functions.lookup(x, [{}],[{}])".format(', '.join([str(col) for col in time_pop]), ', '.join([str(col) for col in vals_pop]))
        updated_params.update({'p_timeseries_pop_growth_rate': new_pop})
    if gdp:
        time_gdp = parsed_excel_sheet.iloc[4, 4:90].tolist()
        vals_gdp = parsed_excel_sheet.iloc[5, 4:90].tolist()
        new_gdp = "functions.lookup(x, [{}],\n\t[{}])".format(', '.join([str(col) for col in time_gdp]), ', '.join([str(col) for col in vals_gdp]))
        updated_params.update({'p_timeseries_gdppc_growth_rate': new_gdp})

    # cells that were updated and that are not constants
    remaining = list(set([cells[0] for cells in cells_updated]) - set(is_cntt))

    return remaining, updated_params


def updated_params_to_file(config, ext):
    """
    Store the new values of the updated params in a file
    :param config: (dict) configuration dictionary
    :param params: (dict) names of params as keys, new values as values
    :param ext: (str) extension of the file
    :return: None
    """
    params = config['update_params']
    file_path_params = _results_naming(config, 'updated_params', ext)
    with open(file_path_params, 'w') as f:
        for func, val in params.items():
            f.write(func + ',' + str(val) + '\n')
            if config['verbose']:
                print('{}: {}'.format(func, val))


def select_model_outputs(config, modelpy):


    # these are the outputs that are plot by default in the plot tool.
    # If they were different for different models, the list could be changed depending
    # on the value of config['region']

    default_outputs = ['tpe_from_res_ej', # total primary energy supply from RES (MToe/Year)
                'total_extraction_nre_ej', # Annual total extraction of non-renewable energy resources (EJ/Year)
                'percent_res_vs_tpes', # Percent of primary energy from RES in the TPES (%)
                'temperature_change', # Temperature of the Atmosphere and Upper Ocean, relative to preindustrial reference period (degreesC)
                'total_land_requirements_renew_mha', # Land required for RES power plants and total bioenergy (land competition + marginal lands (MHa)
                'share_blue_water_use_vs_ar', # Share of blue water used vs accessible runoff water (Dmnl)
                'gdppc',  # GDP per capita (1995T$ per capita) ($/people)
                'eroist_system',  # EROI standard of the system (Dmnl)
                'tfes_intensity_ej_tdollar',# Total final energy intensity (EJ/T$)
                'real_tfec',  # Real total final energy consumption (EJ)
                'gdp',  # Global GDP in T1995T$ (T$)
                'population']  # Population projection (people)


    # the lookups list contains a list of all the functions of the model_py file
    pattern_lookups = re.compile(r"[\w\_]+\(x\)")  # matches all alphanumeric characters followed by '(x)'
    lookups = [x[:-3] for x in pattern_lookups.findall(modelpy)]

    #  make the list of 'step'-cached (i.e. functions of time) variables
    pattern = re.compile(r"@cache\('step'\)\s*def ([\w\_]*)\(\)")

    all_columns = []
    i = 0
    for m in re.finditer(pattern, modelpy):
        funcname = m.group(1).strip()
        if not funcname.startswith('_'):
            i += 1
            all_columns.append(funcname)

    sorted_list = sorted(list(set(all_columns)))

    if config['silent']:
        col_ind = 'r'
    else:
        for num, var_name in enumerate(sorted_list, 1):
            print('{}: {}'.format(num, var_name))

        message ="\n\nPlease select the desired output variables:" \
                 "\n - 0 for all model variables" \
                 "\n - comma separated numbers from the list above for individual variables (e.g.: 1,4,5)" \
                 "\n - r for the same variables of the last simulation" \
                 "\n - + followed by the variable number to add individual variables to the last simulation output (e.g. +5,+19,+21)" \
                 "\n\n Write your choice here:"

        while True:
            col_ind = input(message)
            if col_ind.strip() == '':
                continue

            col_ind_split = [x.strip(' \t\n\b+') for x in col_ind.split(',')]

            if any(x in col_ind_split for x in ['r', '0']):
                break

            col_indices = [int(x) - 1 for x in col_ind_split if x.isdigit()]
            col_vars = [x for x in col_ind_split if not x.isdigit()]

            try:
                return_columns_set = set(list(map(sorted_list.__getitem__, col_indices)) + col_vars)
            except IndexError:
                log.warning('\t\tWrong numerical index, try again...')
                continue

            # check if the variable names are correct
            if return_columns_set.issubset(sorted_list):
                break
            else:
                log.warning('Wrong variable name, try again...')
    col_ind_split = [x.strip(' \t\n\b+') for x in col_ind.split(',')]
    col_indices = [int(x) - 1 for x in col_ind_split if x.isdigit()]
    col_vars = [x for x in col_ind_split if not x.isdigit()]

    return_columns_set = set(list(map(sorted_list.__getitem__, col_indices)) + col_vars)

    if -1 in col_indices:
        return_columns = sorted_list

    elif 'r' in col_vars:
        try:
            with open(os.path.join(config['folder'], 'last_output_conf.txt'), 'r', encoding='utf-8') as f:
                return_columns = sorted([x.strip().split('[')[0] for x in f.readline().split(";")])
        except:
            log.warning("no last configuration, take all vars")
            return_columns = sorted_list
        else:
            log.info('The number and type of output variables of the current '
                     'simulation will be the same as in the previous one.\nIf '
                     'you want to change the number of outputs please remove the'
                     ' "-s" (silent mode) from the options when you run a simulation')

    elif col_ind.lstrip().startswith('+'):
        try:
            with open(os.path.join(config['folder'], 'last_output_conf.txt'), 'r', encoding='utf-8') as f:
                return_columns_set.update([x.strip().split('[')[0] for x in f.readline().split(";")])
        except:
            log.warning("There is no previous simulation available, taking only the given variables instead")
        return_columns = sorted(list(return_columns_set))

    else:
        return_columns = sorted(list(return_columns_set))

    # adding the default variables to the choice of the user
    return_columns = list(set(return_columns + default_outputs))

    with open(os.path.join(config['folder'], 'last_output_conf.txt'), 'w', encoding='utf-8') as f:
        f.write(";".join(return_columns))

    return return_columns


def _chunks(l, n):
    """Yield successive n-sized chunks from a list."""
    for i in range(0, len(l), n):
        yield l[i:i + n]


def create_xarray(data, indexes, model_idxs):
    """
    Create and return xarray DataArray
    :param data: (list) data of the xarray
    :param indexes: (list) list of indexes of the matrix
    :param model_idxs: (dict) all indexes of the model
    :return: (xarray.DataArray) resulting xarray
    """

    if len(indexes) == 1:
        values = data
    elif len(indexes) == 2:
        values = (list(_chunks(data, len(model_idxs[indexes[1]]))))
    elif len(indexes) == 3:
        values = list_2_3d_array(data, len(model_idxs[indexes[0]]), len(model_idxs[indexes[1]]), len(model_idxs[indexes[2]]))
    coords = {}
    for x in indexes:
        coords[x] = model_idxs[x]

    array = xr.DataArray(data=values,
                         coords=coords,
                         dims=indexes)
    return array


def update_matrices(model, config, mod_cells):
    """
    Updates components that are 2D or 3D matrices with the corresponding xarray. It finds the name of the function by
    looking for the name of the modified cell in the range of excel cells in which the function is defined, which are
    specified in the matrices.json file.

    :param model: (pysd.model) model object
    :param config: (dict) configuration dictionary
    :param mod_cells: (list) list of modified excel cell names
    :return: (dict) containing names of variables as keys and xarrays as values
    """
    # load model indexes
    model_idxs = model.components._subscript_dict

    # load excel sheet
    wb = load_workbook(os.path.join(config['folder'], config['inputs_file']),
                       read_only=True, data_only=True)
    ws = wb[config['scenario_sheet'].upper()]

    # loading matrices file
    with open(os.path.join(config['folder'], 'matrices.json'), 'r') as mats:
        m = json.load(mats)

    # transforming cell name into row and col numbers
    mod_xy = map(lambda x: coordinate_from_string(x), mod_cells)

    arrays2mod = {}
    for item in mod_xy:
        col, row = column_index_from_string(item[0]), item[1]

        # getting the values for every
        for name, value in m.items():
            col_0 = value.get('col_0')
            col_n = value.get('col_n')
            if col_0 and col_n:
                col_range = range(col_0 + 1, col_n + 1, 1)
            row_0 = value.get('row_0')
            row_n = value.get('row_n')
            if row_0 and row_n:
                row_range = range(row_0, row_n, 1)
            else:
                row_range = []
            indexes = value.get('indexes')

            # 2D matrix (defined in a single excel row, and several columns)
            if not row_range:
                if (row == row_0) and (col in col_range) and (name not in arrays2mod.keys()):
                    arrays2mod[name] = {'cols': col_range, 'rows': [row_0], 'indexes': indexes}
            # 3D matrix (defined in several excel rows and several columns)
            else:
                if (row in row_range) and (col in col_range) and (name not in arrays2mod.keys()):
                    arrays2mod[name] = {'cols': col_range, 'rows': row_range, 'indexes': indexes}

    new_vals = {}
    for varname, val in arrays2mod.items():
        vals = [ws.cell(row=j, column=i).value for j in val['rows'] for i in val['cols']]
        array = create_xarray(vals, val['indexes'], model_idxs)
        new_vals.update({varname: array})

    return new_vals


def update_model_component(model, comp, new_value):
    """
    Creates a new function to replace the old one using the model.components
    method from pysd
    :param model: model object (pysd.model)
    :parameter: name of the component that needs to be changed
    """

    if isinstance(new_value, str):
        # definition of the new function name
        new_name = "new_{}".format(comp)
        # execution of the new function
        if "lookup" in new_value:
            exec("def {}(x): return {}".format(new_name, new_value))
        else:
            exec("def {}(): return {}".format(new_name, new_value))

        # replacing the old function definition by the new one
        exec("model.components.{} = {}".format(comp, new_name))

    elif isinstance(new_value, (int, float, xr.DataArray)):
        setattr(model.components, comp, lambda: new_value)
    else:
        log.error('Provide a valid data structure')
        sys.exit(0)


def load_model(model_py):
    return pysd.load(os.path.join(model_py))


def run(config, model, run_params, return_columns):
    """

    :param config: (dict) configuration parameters
    :param model: (pysd.model) pysd model object
    :param run_params: (dict) simulation parameters
    :param return_columns: () variables that are to be written in the outputs file
    :return: (pandas.DataFrame) result of the simulation
    """
    print(
        "\n\nSimulation parameters:\n"
        "- Scenario: {}\n"
        "- Initial time: {}\n"
        "- Final time: {}\n"
        "- Time step: {} years ({} days)".format(config['scenario_sheet'].upper(),
                                                 round(run_params['initial_time']),
                                                 round(run_params['final_time']),
                                                 run_params['time_step'],
                                                 round(run_params['time_step']*365))
    )

    if config['update_params']:
        print("- Updated parameters:")
        for key, val in config['update_params'].items():
            if isinstance(val, (int, float)):
                print("\t-" + key + ':' + str(val))
            elif isinstance(val, xr.DataArray):
                print("\t-" + key + ':' + str(val.data.tolist()))
            elif isinstance(val, str):
                if "lookup" in val:
                    print("\t-" + key + ':' + val.lstrip("functions.lookup(x, ").strip(")"))
                else:
                    print("\t-" + key + ':' + val)
        print("\n")
    else:
        print("\n")

    if return_columns[0] != '':
        return_columns = return_columns
    else:
        return_columns = None

    if not config['return_timestep'] is None:
        return_timestamps = np.arange(run_params['initial_time'], run_params['final_time'] + 0.01,
                                      float(config['return_timestep']))
    else:
        return_timestamps = None

    print("Starting simulation.")
    sim_start_time = time.time()
    stocks = model.run(run_params, return_columns=return_columns, return_timestamps=return_timestamps)
    sim_time = time.time() - sim_start_time
    log.info(f"Total simulation time: {(sim_time/60.):.2f} minutes")
    stocks.index.name = 'time'

    return stocks


def list_2_3d_array(data, big_index, col_index, row_index):
    """
    Converts a list that represents a 3D matrix in a 3D array
    e.g. [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12] -> [[[1, 2, 3],[10, 11, 12]],[[4, 5, 6],[13, 14, 15]],[[7, 8, 9],[16,17,18]]]
    :param big_index (int) length of the index that includes the col_index
    :param col_index (int) length of the index that takes several columns
    :param row_index (int) length of the index that takes several columns
    :return: list of lists representing 3D matrix
    """
    mat3d = []
    r = 0
    for i in range(0, big_index, 1):
        int_list = []
        b = r
        for row in range(row_index):
            small_list = data[b:b+row_index]
            int_list.append(small_list)
            b += col_index * big_index

        mat3d.append(int_list)
        r += row_index
    return mat3d
